/*
   This example shows how to create and modify
   configurable rendering pipelines.
*/
import java.io.File;
import java.awt.Color;

public class PipelineExperiment
{
   public static void main(String[] args)
   {
      // Create a Material object (brass).
      Material mtrl1 = new Material();
      mtrl1.ambient[0] = 0.33f;
      mtrl1.ambient[1] = 0.22f;
      mtrl1.ambient[2] = 0.03f;
      mtrl1.ambient[3] = 1.00f;
      mtrl1.diffuse[0] = 0.78f;
      mtrl1.diffuse[1] = 0.57f;
      mtrl1.diffuse[2] = 0.11f;
      mtrl1.diffuse[3] = 1.00f;
      mtrl1.specular[0] = 0.99f;
      mtrl1.specular[1] = 0.91f;
      mtrl1.specular[2] = 0.81f;
      mtrl1.specular[3] = 1.00f;
      mtrl1.s = 27.8f;

      // Create a Material object (chrome).
      Material mtrl2 = new Material();
      mtrl2.ambient[0] = 0.25f;
      mtrl2.ambient[1] = 0.25f;
      mtrl2.ambient[2] = 0.25f;
      mtrl2.ambient[3] = 1.00f;
      mtrl2.diffuse[0] = 0.4f;
      mtrl2.diffuse[1] = 0.4f;
      mtrl2.diffuse[2] = 0.4f;
      mtrl2.diffuse[3] = 1.00f;
      mtrl2.specular[0] = 0.775f;
      mtrl2.specular[1] = 0.775f;
      mtrl2.specular[2] = 0.775f;
      mtrl2.specular[3] = 1.00f;
      mtrl2.s = 76.8f;

      // Create a Light object for the scene.
      Light light = new Light();
      light.pos[0] = 0.0f;
      light.pos[1] = 0.0f;
      light.pos[2] = 10.0f;
      light.pos[3] = 1.0f;
      light.ambient[0] = 1.0f;
      light.ambient[1] = 1.0f;
      light.ambient[2] = 1.0f;
      light.ambient[3] = 1.0f;
      light.diffuse[0] = 1.0f;
      light.diffuse[1] = 1.0f;
      light.diffuse[2] = 1.0f;
      light.diffuse[3] = 1.0f;
      light.specular[0] = 1.0f;
      light.specular[1] = 1.0f;
      light.specular[2] = 1.0f;
      light.specular[3] = 1.0f;


      // Create several Model objects.
      Model m1 = new ObjSimpleModel(new File("resources/horse.obj"));
      for (Triangle t : m1.triangleList)
         t.setMaterial( mtrl1 );
      m1.doLighting = true;
      m1.doSmoothLighting = false;
      m1.doBackFaceCulling = true;
      m1.modelScale(0.5, 0.5, 0.5);
      m1.modelTranslate(-4, 0, -4);
      m1.modelRotate(180, 0,1,0);

      Model m2 = new ObjSimpleModel(new File("resources/small_rhombicosidodecahedron.obj"));
      for (Triangle t : m2.triangleList)
         t.setMaterial( mtrl2 );
      m2.doLighting = true;
      m2.doSmoothLighting = false;
      m2.doBackFaceCulling = true;
      m2.modelScale(0.3, 0.3, 0.3);
      m2.modelTranslate(5, 0, 5);

      Model m3 = new ObjSimpleModel(new File("resources/great_rhombicosidodecahedron.obj"));
      for (Triangle t : m3.triangleList)
         t.setMaterial( mtrl2 );
      m3.doLighting = true;
      m3.doSmoothLighting = false;
      m3.doBackFaceCulling = true;
      m3.modelScale(0.3, 0.3, 0.3);
      m3.modelTranslate(0, 0, 0);

      Model m4 = new ObjSimpleModel(new File("resources/apple.obj"));
      for (Triangle t : m4.triangleList)
         t.setMaterial( mtrl1 );
      m4.doLighting = true;
      m4.doSmoothLighting = false;
      m4.doBackFaceCulling = true;
      m4.modelTranslate(0, 2, 0);
      m4.modelScale(2, 2, 2);
      m4.modelRotate(90, 0,0,1);

      Model m5 = new SphereModel(6, 6, 1);
      for (Triangle t : m5.triangleList)
         t.setMaterial( mtrl1 );
      m5.doLighting = true;
      m5.doSmoothLighting = false;
      m5.doBackFaceCulling = true;
      m5.modelTranslate(-3, 2, 0);

      // Create the Scene object that we shall render
      Scene scene = new Scene();
      // Add the light to the scene.
      scene.light = light;

      // Add the models to the Scene.
      scene.modelList.add(m1);
      scene.modelList.add(m2);
      scene.modelList.add(m3);
      scene.modelList.add(m4);
      scene.modelList.add(m5);

      // Give the Scene a view matrix.
      scene.view2Identity();

      // Give the scene a projection matrix.
      scene.proj2Identity();

      // Set up the scene's view frustum.
      double fov    = 50.0;
      double aspect = 1.0;
      double near   = 1.0;
      double far    = 20.0;
      scene.projPerspective(fov, aspect, near, far);

      // Set up the scene's "camera".
      double eyex=0,    eyey=0,    eyez=10;
      double centerx=0, centery=0, centerz=0;
      double upx=0,     upy=1,     upz=0;
      scene.viewLookAt(eyex,    eyey,    eyez,
                       centerx, centery, centerz,
                       upx,     upy,     upz);


      // Create a framebuffer to render our scene into.
      int vp_width  = 512;
      int vp_height = 512;
      FrameBuffer fb = new FrameBuffer(vp_width, vp_height);
      // Give the framebuffer a nice background color.
      fb.clearFB(java.awt.Color.lightGray);



      // Create a non-standard pipeline.
      SceneRender renderer1 = new SceneRender(scene);
      // put the model transformation stage into this pipeline
      renderer1.addPipelineStage( new P1_Model2World() );
      // put the view transformation stage into this pipeline
      renderer1.addPipelineStage( new P2_World2View(scene) ); // needs the view matrix
      // put the backface culling stage into this pipeline
      renderer1.addPipelineStage( new P3_BackFaceCulling() );
      // put the (flat) vertex shading stage into this pipeline
      renderer1.addPipelineStage( new P4_LightingFlat(scene)   ); // needs lighting information
      // Print out the pipeline.
      System.err.println( renderer1 );
      // Use the pipeline to render our Scene into another Scene.object
      Scene scene2 = renderer1.renderScene();
      // Fix up the new Scene object and its Models.
      scene2.view2Identity();
      for (Model m : scene2.modelList)
      {
         m.model2Identity();
         m.doBackFaceCulling = false;
         m.doLighting = false;
      }

      // Re-render the Scene into the framebuffer using a full pipeline.
      SceneRender renderer2 = new SceneRender(scene2, fb);  // default pipeline
      // Print out this pipeline.
      System.err.println( renderer2 );
      // re-render the scene
      renderer2.renderScene();
      // Save the resulting image in a file.
      fb.dumpFB2File( String.format("PPM_PipelineExperiment_Frame%02d.ppm", 0) );

      // Spin the scene 360 degrees.
      for (int i = 1; i <= 36; i++)
      {
         // Rotate the new scene by 10 degrees.
         scene2.viewTranslate(0, 0, -10);
         scene2.viewRotate(10, 0, 1, 0);
         scene2.viewTranslate(0, 0,  10);
         // Render again.
         fb.clearFB(java.awt.Color.lightGray);
         renderer2.renderScene();
         fb.dumpFB2File(String.format("PPM_PipelineExperiment_Frame%02d.ppm", i));
      }
   }
}