/**
   A Model data structure is mainly a list of Triangle objects
   plus a model matrix (which determines the model's location in
   world coordinates).
*/
import java.util.LinkedList;

public class Model
{
   public Matrix modelMatrix;

   public LinkedList<Triangle> triangleList = new LinkedList<Triangle>();

   public boolean bilerp_flag;  // determines which rendering mode to use for textures.
   public boolean doBackFaceCulling;
   public boolean frontFacingIsCCW;
   public boolean doLighting;
   public boolean doSmoothLighting;
   public boolean doOpenglLighting;  // lighting in the OpenGL style
   public int useVertexNumber;       // should be 0, 1, or 2
   public boolean useHalfVector;
   public boolean doSmoothShading;
   public boolean doTransparancy;


   public Model()
   {
      modelMatrix          = new Matrix(1.0); // identity matrix
      bilerp_flag          = false;
      doBackFaceCulling    = false;
      frontFacingIsCCW     = true;
      doLighting           = false;
      doSmoothLighting     = true;
      doOpenglLighting     = false;
      useVertexNumber      = 0;
      useHalfVector        = false;
      doSmoothShading      = true;
      doTransparancy       = false;
   }// default constructor


   /**
      Duplicate a Model object.
   */
   public Model(Model model)
   {
      // Copy everything except the list of triangles.
      this.modelMatrix          = model.modelMatrix;
      this.bilerp_flag          = model.bilerp_flag;
      this.frontFacingIsCCW     = model.frontFacingIsCCW;
      this.doBackFaceCulling    = model.doBackFaceCulling;
      this.doLighting           = model.doLighting;
      this.doSmoothLighting     = model.doSmoothLighting;
      this.doOpenglLighting     = model.doOpenglLighting;
      this.useVertexNumber      = model.useVertexNumber;
      this.useHalfVector        = model.useHalfVector;
      this.doSmoothShading      = model.doSmoothShading;
      this.doTransparancy       = model.doTransparancy;
   }//duplicating constructor


   public void addTriangle(Triangle t)
   {
      this.triangleList.add(t);
   }


   public void model2Identity()
   {
      modelMatrix = new Matrix(1.0);
   }


   public void modelMult(Matrix m)
   {
      modelMatrix = modelMatrix.times(m);
   }


   public void modelRotate(double theta, double x, double y, double z)
   {
      modelMatrix = modelMatrix.times(new Matrix(theta, x, y, z));
   }


   public void modelScale(double x, double y, double z)
   {
      modelMatrix = modelMatrix.times(new Matrix(x, y, z));
   }


   public void modelTranslate(double x, double y, double z)
   {
      modelMatrix = modelMatrix.times(new Matrix(new Vector(x, y, z)));
   }


   public String toString()
   {
      String result = "";
      result += "This Model has " + triangleList.size() + " triangles\n";;
      result += "doBackFaceCulling = "    + doBackFaceCulling + "\n";
      result += "frontFacingIsCCW = "     + frontFacingIsCCW + "\n";
      result += "doLighting = "           + doLighting + "\n";
      result += "doSmoothLighting = "     + doSmoothLighting + "\n";
      result += "doOpenglLighting = "     + doOpenglLighting + "\n";
      result += "useVertexNumber = "      + useVertexNumber + "\n";
      result += "useHalfVector = "        + useHalfVector + "\n";
      result += "doSmoothShading = "      + doSmoothShading + "\n";
      result += "doTransparancy = "       + doTransparancy + "\n";
      return result;
   }


   /**
      Used for debugging. What this method returns
      depends on the value of the environment variable
      "PrintLevel".

      The value of "PrintLevel" is treated as a bit field.
      The bits have the following meanings.
        1 => print vertex coordinates
        2 => print r g b values
        4 => print texture coordinates
        8 => print normal vector
   */
   public String printVertices()
   {
      String result = "";
      String printString = System.getenv("PrintLevel");
      if (printString != null)
      {
         int vertexFields = Integer.parseInt(printString);

         result += printVertices(vertexFields);
      }
      return result;
   }


   /**
      Used for debugging. What this method returns
      depends on the value of the parameter vertexFields.

      The value of vertexFields is treated as a bit field.
      The bits have the following meanings.
        1 => print vertex coordinates
        2 => print r g b values
        4 => print texture coordinates
        8 => print normal vector
   */
   public String printVertices(int vertexFields)
   {
      String result = toString();
      //result = "Printing out this Model's " + triangleList.size() + " Triangles:\n";
      for (Triangle t : this.triangleList)
      {
         result += t.toString(vertexFields);
      }
      //result += "Done printing out Model\n";
      return result;
   }


   /**
      Used for debugging. What this method returns
      depends on the value of the environment variable
      "MatrixFields".

      The value of "MatrixFields" is treated as a bit field.
      The bits have the following meanings.
        1 => print view matrix
        2 => print projection matrix
        4 => print model matrix
   */
   public String printMatrices()
   {
      String result = "";
      String matrixString = System.getenv("MatrixFields");
      if (matrixString != null)
      {
         int matrixFields = Integer.parseInt(matrixString);

         result += printMatrices(matrixFields);
      }
      return result;
   }


   /**
      Used for debugging. What this method prints out
      depends on the value of the parameter matrixFields.

      The value of matrixFields is treated as a bit field.
      The bits have the following meanings.
        1 => print view matrix
        2 => print projection matrix
        4 => print model matrix
   */
   public String printMatrices(int matrixFields)
   {
      String result = "";

      if ((matrixFields & 4) == 4)
      {
         result += modelMatrix.toString();
      }
      return result;
   }
}