/**

*/
import java.awt.image.BufferedImage;
import java.awt.image.DataBufferByte;
import java.awt.image.WritableRaster;
import java.io.DataInputStream;
import java.io.BufferedInputStream;
import java.io.FileInputStream;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;

import javax.imageio.ImageIO;

public class Texture
{
   public int width;
   public int height;
   public byte[] data; // contains the pixel color data for the texture


   /**
      Create a Texture object directly from pixel data.
   */
   public Texture(int width, int height, byte[] data)
   {
      this.width = width;
      this.height = height;
      this.data = data;
   }


   /**
      Create a Texture object from pixel data in an image file.
   */
   public Texture(String filename)
   {
      try
      {
         DataInputStream dis = new DataInputStream(
                                   new BufferedInputStream(
                                       new FileInputStream(filename)));

         if((Integer.toHexString(dis.readInt()).substring(0, 4)).equals("5036"))
         {
            readPPMTexture(filename);
         }
         else
         {
            readTexture(filename);
         }
      }
      catch(IOException e)
      {
         e.printStackTrace(System.err);
         System.err.println("File not found.");
         System.exit(-1);
      }
   }


   /**
      Read the texture data using Java's image library.

      http://docs.oracle.com/javase/6/docs/api/javax/imageio/ImageIO.html
   */
   public void readTexture(String filename)
   {
      BufferedImage img = null;
      try
      {
         img = ImageIO.read(new File(filename));
      }
      catch(IOException e)
      {
         e.printStackTrace(System.err);
         System.err.println("Cannot read image file.");
         System.exit(-1);
      }

      width = img.getWidth();
      height = img.getHeight();
      WritableRaster raster = img.getRaster();
      DataBufferByte rast_data = (DataBufferByte)raster.getDataBuffer();

      data = rast_data.getData();
   }


   /**
      Read the texture data from pixel data in a PPM file.

   http://stackoverflow.com/questions/2693631/read-ppm-file-and-store-it-in-an-array-coded-with-c
   */
   public void readPPMTexture(String filename) // name of the ppm file to load texture data from
   {
      FileInputStream fis = null;
      try
      {
         fis = new FileInputStream(filename);
      }
      catch (FileNotFoundException e)
      {
         e.printStackTrace(System.err);
         System.err.printf("ERROR! Could not open texture %s\n", filename);
         System.exit(-1);
      }

      try
      {
         // Read image format string "P6".
         String magicNumber = "";
         char c = (char)fis.read();
         while (c != '\n')
         {
            magicNumber += c;
            c = (char)fis.read();
         }
         if (! magicNumber.trim().startsWith("P6"))
         {
            System.err.printf("ERROR! Improper PPM number in file %s\n", filename);
            System.exit(-1);
         }

         c = (char)fis.read();
         if ( '#' == c ) // read (and discard) IrfanView comment
         {
            while (c != '\n')
            {
               c = (char)fis.read();
            }
            c = (char)fis.read();
         }

         // Read image dimensions.
         String widthDim = "";
         while (c != ' ' && c != '\n')
         {
            widthDim += c;
            c = (char)fis.read();
         }
         width = Integer.parseInt(widthDim.trim());

         String heightDim = "";
         c = (char)fis.read();
         while (c != '\n')
         {
            heightDim += c;
            c = (char)fis.read();
         }
         height = Integer.parseInt(heightDim.trim());
         System.err.printf("Texture %s: width=%d, height=%d\n\n",filename,width,height);

         // Read image rgb dimensions (which we don't use).
         c = (char)fis.read();
         while (c != '\n')
         {
            c = (char)fis.read();
         }

         // Create the data array.
         data = new byte[width*height*3];
         // Read pixel data from the PPM file.

         // Read one row of pixels at a time.
         for (int y = 0; y < height; y++)
         {  // Read into the bottom row of the data buffer
            // up towards the top row.
            if ( fis.read(data, (height-y-1)*3*width, 3*width) != 3*width )
            {
               System.err.printf("ERROR! Could not load texture %s\n", filename);
               System.exit(-1);
            }
         }
/*
         // Read the entire texture into the data buffer in one call to read().
         // This makes the texture upside down.
         if ( fis.read(data, 0, 3*width*height) != 3*width*height )
         {
            System.err.printf("ERROR! Could not load texture %s\n", filename);
            System.exit(-1);
         }
*/
         fis.close();
      }
      catch (IOException e)
      {
         e.printStackTrace(System.err);
         System.err.printf("ERROR! Could not read texture %s\n", filename);
         System.exit(-1);
      }
   }//readPPMTexture()


   /**
      Simple test of this Texture class.
   */
   public static void main(String[] args)
   {
      Texture t1 = new Texture("../Clients/Resources/Yoda.ppm");
      Texture t2 = new Texture("../Clients/Resources/Yoda.png");
      Texture t3 = new Texture("../Clients/Resources/Yoda.jpg");
      Texture t4 = new Texture("../Clients/Resources/Yoda.gif");  // not supported?

      Texture t = t4;

      javax.swing.JFrame frame = new javax.swing.JFrame( "Drawing 2D Shapes" );
      frame.setDefaultCloseOperation( javax.swing.JFrame.EXIT_ON_CLOSE );
      frame.setSize(t.width+50, t.height+100);
      frame.setVisible( true );
      java.awt.Graphics gr = frame.getGraphics();

      java.awt.Image image = frame.createImage(t.width+50, t.height+100);
      java.awt.Graphics gOff = image.getGraphics();
      gOff.setColor( java.awt.Color.black );
      gOff.fillRect(0, 0, t.width+50, t.height+100);
      for (int y = 0; y < t.height; y++)
         for (int x = 0; x < t.width; x++)
         {
            int r = t.data[(y*t.width + x)*3 + 0];
            int g = t.data[(y*t.width + x)*3 + 1];
            int b = t.data[(y*t.width + x)*3 + 2];
            if (r < 0) r = 256+r;
            if (g < 0) g = 256+g;
            if (b < 0) b = 256+b;
            gOff.setColor( new java.awt.Color(r, g, b) );
            gOff.drawLine(x+25, y+50, x+25, y+50);  // plot one pixel
         }
/*
      java.awt.image.DataBuffer db
         = new java.awt.image.DataBufferByte(t.data,
                                             t.width * t.height * 3);
      java.awt.image.Raster raster
         = java.awt.image.Raster.createInterleavedRaster(db,
                                                         t.width,
                                                         t.height,
                                                         t.width * 3,
                                                         3,
                                                         new int[]{0, 1, 2},
                                                         null);
      java.awt.image.ColorModel cm
         = new java.awt.image.ComponentColorModel(java.awt.color.ColorSpace.getInstance(java.awt.color.ColorSpace.CS_sRGB),
                                                  new int[]{8, 8, 8},
                                                  false,
                                                  false,
                                                  java.awt.Transparency.OPAQUE,
                                                  java.awt.image.DataBuffer.TYPE_BYTE);
      java.awt.Image image
         =  new java.awt.image.BufferedImage(cm, raster, false, null);
*/
      gr.drawImage(image, 0, 0, null);
   }//main()
}//Texture