/**
   Shade each fragment created by the rasterizer.
   If a Triangle has a Texture, use the texture
   to shade the triangle's fragments.
*/

public class P10_FragmentShaderTexture implements PipelineStage
{
   public P10_FragmentShaderTexture()
   {
   }

   /**
      Walk the Model's list of Triangle objects and for each triangle,
      walk its list of Fragment objects. For each fragment, use
      the barycentric coordinates in the fragment to interpolate
      texture data from the triangle's three vertices to the fragment.
   */
   public Model processModel(Model model)
   {
      int triangleCounter = 0;  // useful for debugging
      for (Triangle triangle : model.triangleList)
      {
         triangleCounter++;
         if (null != triangle.tex)  // if this triangle has a texture
         {
            // For this triangle, get the texture coordinates
            // from each of the three vertices.
            double s0 = triangle.v[0].s;
            double t0 = triangle.v[0].t;
            double s1 = triangle.v[1].s;
            double t1 = triangle.v[1].t;
            double s2 = triangle.v[2].s;
            double t2 = triangle.v[2].t;

            // Walk this triangle's list of fragments.
            for (Fragment frag : triangle.fragments)
            {
               // get this fragment's barycentric coordinates
               double alpha = frag.alpha;
               double beta  = frag.beta;
               double gamma = frag.gamma;

               //System.err.printf("P10 alpha = %f, beta = %f, gamma = %f\n", alpha, beta, gamma);

               // interpolate texture coordinates, for this fragment, from the triangle's vertices
               double s = alpha*s0 + beta*s1 + gamma*s2;
               double t = alpha*t0 + beta*t1 + gamma*t2;

               s = s - Math.floor(s);  // wrap the texture in both the
               t = t - Math.floor(t);  // s and t coordinate directions
                                       // for both pos and neg texture coordinates

               //System.err.printf("P10 s = %f, t = %f\n", frag.s, frag.t);

               // convert (interpolated) texture coordinates into pixel indices
               int y = (int)(t * triangle.tex.height);  // row index
               int x = (int)(s * triangle.tex.width);   // column index

               // copy color data from the texture to this fragment
               int index = ((y * triangle.tex.width) + x) * 3;
               int r = triangle.tex.data[index + 0];
               int g = triangle.tex.data[index + 1];
               int b = triangle.tex.data[index + 2];
               if (r < 0) r = 256+r;
               if (g < 0) g = 256+g;
               if (b < 0) b = 256+b;
               frag.r = r / 255.0;
               frag.g = g / 255.0;
               frag.b = b / 255.0;

               frag.a = 1.0;
               //System.err.printf("P10 r = %f, g = %f, b = %f\n", frag.r, frag.g, frag.b);
            }
         }
      }
      return model;
   }
}