/**
   Compute the shade (i.e., color) of each vertex
   as determined by all the lighting data.
*/

public class P4_LightingSmooth implements PipelineStage
{
   Scene scene;

   public P4_LightingSmooth(Scene scene)
   {
      this.scene = scene;
   }


   /**
      Walk the Model's list of Triangle objects.
      For each Vertex in each Triangle object, use
      the Scene's light data, the Triangle's material
      properties, and the Vertex's normal vector to
      compute color information for that vertex.
   */
   public Model processModel(Model model)
   {
      Light light = scene.light;
      Matrix viewMatrix = scene.viewMatrix;

      // check if we should do smooth lighting
      if ( !model.doLighting || !model.doSmoothLighting )
      {
         return model;
      }

      if ( null == light )
      {
         System.err.println("Warning! Lighting is enabled but the Scene does not have a Light.");
         return model;
      }

      // Here is the light source as a vertex in world(!) coordinates.
      Vertex S = new Vertex(light.pos);
      // Convert light source from world coordinates to view coordinates.
      S = viewMatrix.times(S);
      // Convert the vertex to a vector.
      Vector S2 = new Vector(S.x, S.y, S.z);

      for (Triangle t : model.triangleList)
      {
         if (null == t.mtrl)
         {
            System.err.println("Warning! Lighting is enabled but a triangle does not have a Material.");
            continue;
         }

         if ( (null == t.v[0].n)
           || (null == t.v[1].n)
           || (null == t.v[2].n) )
         {
            System.err.println("Warning! Smooth lighting is enabled but a vertex does not have a normal vector.");
            continue;
         }

         // Get all of the values that we need.
         double s = t.mtrl.s; //shininess coefficient
         double kar = t.mtrl.ambient[0];  // red
         double kdr = t.mtrl.diffuse[0];
         double ksr = t.mtrl.specular[0];
         double car = light.ambient[0];
         double cdr = light.diffuse[0];
         double csr = light.specular[0];
         double kag = t.mtrl.ambient[1];  // green
         double kdg = t.mtrl.diffuse[1];
         double ksg = t.mtrl.specular[1];
         double cag = light.ambient[1];
         double cdg = light.diffuse[1];
         double csg = light.specular[1];
         double kab = t.mtrl.ambient[2];  // blue
         double kdb = t.mtrl.diffuse[2];
         double ksb = t.mtrl.specular[2];
         double cab = light.ambient[2];
         double cdb = light.diffuse[2];
         double csb = light.specular[2];

         // Compute a color value at each of the three vertices of this triangle.
         for (int i = 0; i < 3; i++)
         {
            // To do vector algebra, we need the vertex as a vector.
            Vector p = new Vector(t.v[i].x, t.v[i].y, t.v[i].z);
            // We need the normal vector at this vertex.
            // Here is this vertex's normal vector in view(!) coordinates.
            Vector N = t.v[i].n;
            // We need the unit light vector, L, from our vertex to the light source.
            // Compute L = S2 - p, the vector from our vertex to the light source.
            Vector L = S2.plus( p.times(-1.0) );
            // normalize L
            L = L.normalize();
            // We need the reflection vector R = 2(L dot N)*N - L.
            Vector R = N.times( 2*L.dotProduct(N) ).plus( L.times(-1.0) );
            // We need the unit view vector, V, that points from the vertex to the origin.
            Vector V = p.normalize().times(-1.0);
            // We can also use the "half vector", H = normalize(L + V).
            Vector H = L.plus(V).normalize();

            // Compute the color at this vertex using its normal, light, reflection, and view vectors,
            //    c = k_a * c_a + sum_{all lights}( k_d*c_d*(L dot N) + k_s*c_s*(R dot V)^s ) )
            // or, using the half vector H,
            //    c = k_a * c_a + sum_{all lights}( k_d*c_d*(L dot N) + k_s*c_s*(N dot H)^s ) )

            // pre compute some numbers
            double diffuseAngle = L.dotProduct(N);
            diffuseAngle = (diffuseAngle < 0) ? 0.0 : diffuseAngle;
            double shininess;
            if (model.useHalfVector)
            {
               shininess = N.dotProduct(H);
            }
            else
            {
               shininess = R.dotProduct(V);
            }
            shininess = (shininess < 0) ? 0.0 : Math.pow(shininess, s);

            // do red
            double r = kar*car + kdr*cdr*diffuseAngle + ksr*csr*shininess;
            r = (r <= 1.0) ? r : 1.0;  // clamp the color value at 1.0
            t.v[i].r = r;

            // do green
            double g = kag*cag + kdg*cdg*diffuseAngle + ksg*csg*shininess;
            g = (g <= 1.0) ? g : 1.0;  // clamp the color value at 1.0
            t.v[i].g = g;

            // do blue
            double b = kab*cab + kdb*cdb*diffuseAngle + ksb*csb*shininess;
            b = (b <= 1.0) ? b : 1.0;  // clamp the color value at 1.0
            t.v[i].b = b;

            if ( model.doOpenglLighting && (i == model.useVertexNumber) )
            {  /* Do OpenGL style lighting. */
               /* Copy the lighting information from vertex v[i]
                  to the other two vertices, v[i+1] and v[i+2]. */
               int j = i + 1;
               if (3 == j) j = 0;
               t.v[j].r = r;
               t.v[j].g = g;
               t.v[j].b = b;
               j++;
               if (3 == j) j = 0;
               t.v[j].r = r;
               t.v[j].g = g;
               t.v[j].b = b;
               break;   /* ugly, ugly, ugly */
            }
         }
      }
      return model;
   }
}