/**
   Convert vertices from normalized device coordinates to viewport coordinates.
*/

public class P8_NDC2VP implements PipelineStage
{
   FrameBuffer fb;

   public P8_NDC2VP(FrameBuffer fb)
   {
      this.fb = fb;
   }


   /**
      Walk the Model's list of Triangle objects and transform
      the coordinates of each vertex, which are now in
      "Normalized Device Coordinates", to "pixel
      coordinates" in the viewport. This is the last
      transformation before we rasterize the triangles.
      This is also the first stage of the pipeline that
      needs to know about the framebuffer.

      The "normalized device coordinates" are all between -1 and +1
      so
           -1 <= x_ndc <= +1
           -1 <= y_ndc <= +1.
      The horizontal viewport coordinates should be between 0 and vpWidth.
      and the vertical viewport coordinates should be between 0 and vpHeight
      so
            0 <= x_vp <= vpWidth
            0 <= y_vp <= vpHeight.
      Notice that viewport coordinates are still floating point numbers (they
      are not yet integer coordinates; that's done in the rasterizer).

      For each vertex in the triangle,
         x_vp = (vpWidth /2.0)*(x_ndc + 1.0)
         y_vp = (vpHeight/2.0)*(y_ndc + 1.0)
         z_vp = z_ndc  (the z-buffer is just the normalized device coordinate)
                       (-1 is the "front" of the z-buffer and +1 is the "back")

      QUESTION: Why don't we skip this stage and rasterize in the
                normalized device coordinates? It would save time,
                and this stage seems very unnecessary.
   */
   public Model processModel(Model model)
   {
      int vpWidth  = fb.getWidthVP();
      int vpHeight = fb.getHeightVP();

      for (Triangle t : model.triangleList)
      {
         // The "normalized device coordinates" are all between -1 and +1.
         // The horizontal viewport coordinates should be between 0 and vpWidth.
         // The vertical viewport coordinates should be between 0 and vpHeight.
         // So x_vp = (vpWidth /2.0)*(x_ndc + 1.0)
         //    y_vp = (vpHeight/2.0)*(y_ndc + 1.0)
         //    z_vp = z_ndc  (the z-buffer is the normalized device coordinate)
         //                  (-1 is the "front" of the z-buffer and +1 is the "back")
         t.v[0].x = (vpWidth /2.0) * (t.v[0].x + 1.0);
         t.v[0].y = (vpHeight/2.0) * (t.v[0].y + 1.0);
         t.v[1].x = (vpWidth /2.0) * (t.v[1].x + 1.0);
         t.v[1].y = (vpHeight/2.0) * (t.v[1].y + 1.0);
         t.v[2].x = (vpWidth /2.0) * (t.v[2].x + 1.0);
         t.v[2].y = (vpHeight/2.0) * (t.v[2].y + 1.0);
         //debug(t);
      }
      return model;
   }


   private static void debug(Triangle t)
   {
      // print the (x,y) part of each vertex in the triangle
      System.err.printf("x=% 6.1f, y=% 6.1f, x=% 6.1f, y=% 6.1f, x=% 6.1f, y=% 6.1f\n",
                         t.v[0].x, t.v[0].y, t.v[1].x, t.v[1].y, t.v[2].x, t.v[2].y);
   }
}