/**
   Rasterize each triangle into a list of fragments.
*/
import java.util.LinkedList;

public class P9_Rasterize implements PipelineStage
{
   public P9_Rasterize()
   {
   }


   /**
      Walk the Model's list of Triangle objects and rasterize each triangle.

      This rasterization algorithm is from
        "Fundamentals of Computer Graphics", 2nd Edition,
         by Peter Shirley, pages 63-66.
   */
   public Model processModel(Model model)
   {
      int triangleCounter = 0;  // useful for debugging
      for (Triangle t : model.triangleList)
      {
         triangleCounter++;

         // Give this triangle a fragment list.
         t.fragments = new LinkedList<Fragment>();

         double x0 = t.v[0].x;
         double y0 = t.v[0].y;
         double x1 = t.v[1].x;
         double y1 = t.v[1].y;
         double x2 = t.v[2].x;
         double y2 = t.v[2].y;

         // Cull degenerate triangles (triangles that projected onto a line).
         // NOTE: This culling must be done after the projection stage.
         //       In view coordinates we can determine front and back
         //       facing triangles, but we can't find the ones that
         //       project to lines until we have done the projection.
         double area = (x1-x0)*(y2-y0) - (y1-y0)*(x2-x0);
         if ( 0.0001 > Math.abs(area) )
            continue;

         // find the greatest integer less than or equal to x0, x1, and x2
         int xMin = (int)x0;
         if (x1 < xMin) xMin = (int)x1;
         if (x2 < xMin) xMin = (int)x2;
//       int xMin = (int)Math.floor(x0);
//       if (x1 < xMin) xMin = (int)Math.floor(x1);
//       if (x2 < xMin) xMin = (int)Math.floor(x2);

         // find the greatest integer less than or equal to y0, y1, and y2
         int yMin = (int)y0;
         if (y1 < yMin) yMin = (int)y1;
         if (y2 < yMin) yMin = (int)y2;

         // find the least integer greater than or equal to x0, x1, and x2
         int xMax = (int)Math.ceil(x0);
         if (x1 > xMax) xMax = (int)Math.ceil(x1);
         if (x2 > xMax) xMax = (int)Math.ceil(x2);
//       int xMax = (int)x0;                   // the next three lines are not correct
//       if (x1 > (xMax + 1)) xMax = (int)x1;
//       if (x2 > (xMax + 1)) xMax = (int)x2;
//       int xMax = ((int)x0) + 1);            // the next three lines are not correct
//       if (x1 > xMax) xMax = ((int)x1) + 1;
//       if (x2 > xMax) xMax = ((int)x2) + 1;

         // find the least integer greater than or equal to y0, y1, and y2
         int yMax = (int)Math.ceil(y0);
         if (y1 > yMax) yMax = (int)Math.ceil(y1);
         if (y2 > yMax) yMax = (int)Math.ceil(y2);

         //System.err.printf("P9 xMin=% 4d, xMax=% 4d, yMin=% 4d, yMax=% 4d\n", xMin,xMax,yMin,yMax);

       //for (int y = yMin; y < yMax; y++) // bottom to top
         for (int y = yMax-1; y >= yMin; y--) // top to bottom
         {
            for (int x = xMin; x < xMax; x++)  /// left to right
            {
               //  f01(x,y) = (y0-y1)*(x) + (x1-x0)*(y) + x0*y1 - x1*y0
               //  f12(x,y) = (y1-y2)*(x) + (x2-x1)*(y) + x1*y2 - x2*y1
               //  f20(x,y) = (y2-y0)*(x) + (x0-x2)*(y) + x2*y0 - x0*y2

               //  alpha = f12(x,y) / f12(x0,y0)
               //  beta  = f20(x,y) / f20(x1,y1)
               //  gamma = f01(x,y) / f01(x2,y2)

               double alpha = ( (y1-y2)*(x)  + (x2-x1)*(y)  + x1*y2 - x2*y1 )
                            / ( (y1-y2)*(x0) + (x2-x1)*(y0) + x1*y2 - x2*y1 );
               double beta  = ( (y2-y0)*(x)  + (x0-x2)*(y)  + x2*y0 - x0*y2 )
                            / ( (y2-y0)*(x1) + (x0-x2)*(y1) + x2*y0 - x0*y2 );
               double gamma = ( (y0-y1)*(x)  + (x1-x0)*(y)  + x0*y1 - x1*y0 )
                            / ( (y0-y1)*(x2) + (x1-x0)*(y2) + x0*y1 - x1*y0 );


               //System.err.printf("P9 alpha = %f, beta = %f, gamma = %f\n", alpha,beta,gamma);

               if (alpha >= 0 && beta >= 0 && gamma >= 0)
               {  // create a new fragment
                  Fragment frag = new Fragment();
                  frag.xIndex = x;
                  frag.yIndex = y;
                  frag.zCoord = alpha*(t.v[0].z) + beta*(t.v[1].z) + gamma*(t.v[2].z);
                  frag.alpha = alpha;
                  frag.beta = beta;
                  frag.gamma = gamma;
                  frag.r = 0;
                  frag.g = 0;
                  frag.b = 0;
                  frag.a = 0;
                  // add this fragment to this triangle's list of fragments
                  t.fragments.add(frag);
                  //System.err.printf("P9 x=%d,y=%d,alpha=%f,beta=%f,gamma=%f\n", x,y,alpha,beta,gamma);
               }
            }
         }
      }
      return model;
   }
}