/**
   This class implements the "rendering pipeline".
*/
import java.util.Vector;

public class SceneRender
{
   private Vector<PipelineStage> pipeline;
   private Scene scene;
   private FrameBuffer fb;

   /**
      This constructor is for specialized renderers
      that do not do all of the usual rendering stages
      (and don't need a framebuffer). Such a renderer
      can fill this pipeline with whatever stages are
      desired.

      Notice that the renderScene() method below returns
      to its caller the result of the last stage of this
      pipeline. So a renderer built with this constructor
      can be thought of as a "render-to-Scene" renderer.
   */
   public SceneRender(Scene scene)
   {
      this.pipeline = new Vector<PipelineStage>();
      this.scene = scene;
   }


   /**
      This constructor initializes the default
      rendering pipeline. This pipeline is modeled
      on the (older) OpenGL fixed function pipeline.

      A client program can replace, delete, or insert
      other stages into this pipeline.
   */
   public SceneRender(Scene scene, FrameBuffer fb)
   {
      this.pipeline = new Vector<PipelineStage>();
      this.scene = scene;
      this.fb = fb;

      // Build the "fixed function" pipeline.

      // duplicate the model
      pipeline.add( new P0_DuplicateModel() );
      // apply the model transformation
      pipeline.add( new P1_Model2World() );
      // apply the view transformation
      pipeline.add( new P2_World2View(scene) );    // needs the view matrix
      // delete back facing triangles
      pipeline.add( new P3_BackFaceCulling() );
      // compute vertex shading
      pipeline.add( new P4_LightingFlat(scene) );  // needs lighting information
      pipeline.add( new P4_LightingSmooth(scene) );
      // apply the projection transformation
      pipeline.add( new P5_View2Clip(scene) );     // needs the projection matrix
      // clip triangles to the standard volume
      pipeline.add( new P6_Clipping() );
      // transform to normalized device coordinates
      pipeline.add( new P7_Clip2NDC() );
      // transform normalized device coordinates to viewport coordinates
      pipeline.add( new P8_NDC2VP(fb) );           // needs the viewport dimensions
      // rasterize each triangle into fragments (why not rasterize in NDC?)
      pipeline.add( new P9_Rasterize() );
      // shade fragments of each triangle
      pipeline.add( new P10_FragmentShaderColor() );
      pipeline.add( new P10_FragmentShaderTexture() );
      // use the fragments to fill in the viewport
      pipeline.add( new P11_Fragments2FB(fb) );    // needs the framebuffer
   }


   /**
      This method renders the Scene object by
      passing each of its Model objects through
      the rendering pipeline and storing the
      processed models in a new Scene object.

      At the end of the pipeline, the resulting
      new Scene object gets passed back to the caller.
      This give a "render-to-Scene" capability for
      some interesting special effects.

      For the default pipeline, at the end of the
      pipeline the Framebuffer's viewport holds the
      rendered image of the scene represented by the
      Scene object.
   */
   public Scene renderScene()
   {
      System.err.print( scene.printMatrices() ); // output depends on an environment variable

      Scene renderedScene = new Scene(scene); // copy everything but the geometry

      int modelNumber = 1;    // useful for debugging
      for (Model model : scene.modelList)
      {
       //System.err.println("Model no. "+modelNumber+": "+model.getClass().getName());
         System.err.print( model.printMatrices() ); // output depends on an environment variable

         // Send this model down the pipeline.
         int stageNumber = 0;    // useful for debugging
         for (PipelineStage ps : pipeline)
         {
            //System.err.println("  Stage no. "+stageNumber+": "+ps.getClass().getName());

            model = ps.processModel(model);

            printVertices(model, stageNumber); // output depends on environment variables

            stageNumber++;
         }

         renderedScene.addModel(model);
         modelNumber++;
      }
      return renderedScene;
   }//renderScene()



   /**
      This method appends a PipelineStage object
      at the end of the current pipeline.
   */
   public void addPipelineStage(PipelineStage ps)
   {
      pipeline.add(ps);
   }


   /**
      This method replaces the pipeline stage at the
      given index with the new PipelineStage object.
   */
   public void replacePipelineStage(int index, PipelineStage ps)
   {
      pipeline.set(index, ps);
   }


   /**
      This method inserts the new pipeline stage at the
      given index. The current PipelineStage object at
      that index, and all subsequent pipeline stage
      objects, shift up in the pipeline.
   */
   public void insertPipelineStage(int index, PipelineStage ps)
   {
      pipeline.add(index, ps);
   }


   /**
      This method deletes the pipeline stage at the
      given index. All subsequent pipeline stage objects
      shift down in the pipeline.
   */
   public PipelineStage removePipelineStage(int index)
   {
      return pipeline.remove(index);
   }


   public String toString()
   {
      String result = "";
      for (PipelineStage ps : pipeline)
      {
         result += ps.getClass().getName() + "\n";
      }
      return result;
   }


   /**
      Used for debugging. What this method prints out
      depends on the value of the environment variables
      "PrintLevel" and "Pipeline*", where * is any positive integer.
   */
   public static void printVertices(Model model, int stageNo)
   {
      int stageFlag = 1;
      for (int i = 1; i < stageNo; i++)
         stageFlag = stageFlag << 1;  // shift the 1 bit into the correct position

      String debugString = null;
      int debugStages = 0;
      // check which (of the first 16) pipeline stages should be printed out
      if ( (debugString = System.getenv("PipeLine1") ) != null ) debugStages += 1;
      if ( (debugString = System.getenv("PipeLine2") ) != null ) debugStages += 2;
      if ( (debugString = System.getenv("PipeLine3") ) != null ) debugStages += 4;
      if ( (debugString = System.getenv("PipeLine4") ) != null ) debugStages += 8;
      if ( (debugString = System.getenv("PipeLine5") ) != null ) debugStages += 16;
      if ( (debugString = System.getenv("PipeLine6") ) != null ) debugStages += 32;
      if ( (debugString = System.getenv("PipeLine7") ) != null ) debugStages += 64;
      if ( (debugString = System.getenv("PipeLine8") ) != null ) debugStages += 128;
      if ( (debugString = System.getenv("PipeLine9") ) != null ) debugStages += 256;
      if ( (debugString = System.getenv("PipeLine10")) != null ) debugStages += 512;
      if ( (debugString = System.getenv("PipeLine11")) != null ) debugStages += 1024;
      if ( (debugString = System.getenv("PipeLine12")) != null ) debugStages += 2048;
      if ( (debugString = System.getenv("PipeLine13")) != null ) debugStages += 4096;
      if ( (debugString = System.getenv("PipeLine14")) != null ) debugStages += 8192;
      if ( (debugString = System.getenv("PipeLine15")) != null ) debugStages += 16384;
      if ( (debugString = System.getenv("PipeLine16")) != null ) debugStages += 32768;

      // do a bitwise-and to determine which stages get printed out
      if ((stageFlag & debugStages) != 0)
      {
         System.err.printf("================= Stage P%d begin ===========\n", stageNo);
         System.err.print( model.printVertices() );
         System.err.printf("================= Stage P%d end =============\n", stageNo);
      }
   }//printVertices()
}